import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
import torch
import math


class clipEstimator(Function):
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        out = torch.sign(input)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        mask = torch.ones_like(input) * (torch.abs(input) <= 1)
        return grad_output * mask


class clipConv2dW1A1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(clipConv2dW1A1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = clipEstimator().apply(w)
        ba = clipEstimator().apply(input)

        output = F.conv2d(ba, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class clipConv2dW1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(clipConv2dW1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = clipEstimator().apply(w)

        output = F.conv2d(input, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class secondOrderEstimator(Function): # implementing BiRealNet paper approximation
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        out = torch.sign(input)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        mask1 = (input >= -1)*1
        mask2 = (input >= 0)*1
        mask3 = (input >= 1)*1

        grad_input = (mask1-mask2)*(2 + 2*input) + (mask2-mask3)*(2-2*input)
        return grad_output*grad_input


class secondOrderConv2dW1A1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(secondOrderConv2dW1A1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = secondOrderEstimator().apply(w)
        ba = secondOrderEstimator().apply(input)

        output = F.conv2d(ba, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class secondOrderConv2dW1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(secondOrderConv2dW1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = secondOrderEstimator().apply(w)

        output = F.conv2d(input, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class bcEstimator(Function): # approx with (2/coshx) ^2 like in binaryConnect?
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        out = torch.sign(input)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        grad_input = (2/torch.cosh(input))**2
        return grad_output*grad_input


class bcConv2dW1A1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(bcConv2dW1A1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = bcEstimator().apply(w)
        ba = bcEstimator().apply(input)

        output = F.conv2d(ba, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class bcConv2dW1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(bcConv2dW1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = bcEstimator().apply(w)

        output = F.conv2d(input, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class irEstimator(Function): # approx with (2/coshx) ^2 like in binaryConnect?
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        out = torch.sign(input)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        grad_input = 1 - torch.pow(torch.tanh(input), 2)
        return grad_output*grad_input


class irConv2dW1A1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(irConv2dW1A1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = irEstimator().apply(w)
        ba = irEstimator().apply(input)

        output = F.conv2d(ba, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output


class irConv2dW1(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True):
        super(irConv2dW1, self).__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias)

    def forward(self, input):
        w = self.weight
        bw = irEstimator().apply(w)

        output = F.conv2d(input, bw, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return output

