import argparse
import os
import sys
import time
import math
import random
import numpy as np
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim
import torch.utils.data
import torchvision.transforms as transforms
import torchvision.datasets as datasets
from models import blitz
from training_parameters import losses
from datetime import datetime
from torchsummary import summary
from tools.useful_tools import ensure_dir
from models import binaryConnectModel, largeBinaryNets
from prettytable import PrettyTable


model_names = sorted(['blitz0', "blitzW1A1", "blitzW1", "blitzGumbelSoftmax", "blitzGumbelSoftmaxW1A1", "blitzSTEW1A1", "blitzSTEW1", "blitzBiRealNet",
           "blitzgclipW1", "blitzgclipW1A1", "blitzgSecondOW1", "blitzgSecondOW1A1", "blitzgBCW1", "blitzgBCW1A1", "blitzgirW1","blitzgirW1A1",
                      "binaryconnect0","binaryconnectW1" ,"bnnA5BIR", "bnnA5B","bnnA10B", "bnnA50B", "bnnA1BK3", "bnnABK3", "bnnF256", "bnnF256gumbel",
                      "vgg_small_1w1a_gumbel", "vgg_small_1w_gumbel"])

loss_choices = sorted(name for name in losses.__dict__
                     if name.islower() and not name.startswith("__")
                     and callable(losses.__dict__[name]))

print("MODEL NAMES", model_names)

parser = argparse.ArgumentParser(description='PyTorch Cifar10 Training')
parser.add_argument('--arch', '-a', metavar='ARCH', default='blitz0', choices=model_names, help='model architecture: ' + ' | '.join(model_names) +' (default: blitz)')
parser.add_argument('--loss', default='crossentropyloss', choices=loss_choices, help="choose which loss criterion to use")
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N', help='number of data loading workers (default: 4)')
parser.add_argument('--c_s', type=float, default=2000., help='for softmax')
parser.add_argument('--method', type=str, default='gumbel-softmax', help='Method; naive, ste, ir-net, gumbel-softmax')
parser.add_argument('--tau_min', type=float, default=1.5)  # 0.5, help='minimum temperature parameter for softmax') # 0.8
parser.add_argument('--tau_max', type=float, default=6.0)  # 3.0, help='maximum temperature parameter for softmax') # 6.0
parser.add_argument('--r', type=float, default=1e-4, help='coefficient for decreasing tau param, deprecated.')
parser.add_argument('--c_scale', type=float, default=300.0)
parser.add_argument('--bias', type=bool, default=False, help='bias in the linear layers?')
parser.add_argument('--epochs', default=300, type=int, metavar='N', help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N', help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=128, type=int, metavar='N', help='mini-batch size (default: 128)')
parser.add_argument('--lr', '--learning-rate', default=0.007, type=float, metavar='LR', help='initial learning rate')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M', help='momentum')
parser.add_argument('--weight-decay', '--wd', default=1e-4, type=float, metavar='W', help='weight decay (default: 1e-4)')
parser.add_argument('--print-freq', '-p', default=40, type=int, metavar='N', help='print frequency (default: 20)')
parser.add_argument('--resume', default='', type=str, metavar='PATH', help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true', help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true', help='use pre-trained model')
parser.add_argument('--half', dest='half', action='store_true', help='use half-precision(16-bit) ')
parser.add_argument('--save-dir', dest='save_dir', help='The directory used to save the trained models', default='save_temp', type=str)
parser.add_argument('--save-every', dest='save_every', help='Saves checkpoints at every specified number of epochs', type=int, default=20)
parser.add_argument('--parameter_count', help='count num of parameters', type=bool, default=False)

best_prec1 = 0


def main():
    global args, best_prec1
    args = parser.parse_args()

    # Check the save_dir exists or not
    # print(str(datetime.now()))
    if not args.parameter_count:
        save_runs_dir = os.path.join(args.save_dir, str(args.arch), "bias"+str(args.bias)+"date" + str(datetime.now()))
        ensure_dir(save_runs_dir)
    # sys.exit("test")

    # txt file to print architecture and commands
    # txtfile = open(os.path.join(save_runs_dir, 'parameters.txt'), "a")
    # print("ARGS", args, file=txtfile)

    if args.arch=='binaryconnect0' or args.arch=='binaryconnectW1':
        model = torch.nn.DataParallel(binaryConnectModel.__dict__[args.arch]())
    # if args.arch =='bnnA_10_B':
    #     model = torch.nn.DataParallel(largeBinaryNets.__dict__[args.arch]())
    else:
        model = torch.nn.DataParallel(blitz.__dict__[args.arch]())

    if not args.parameter_count:
        model.cuda()

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            checkpoint = torch.load(args.resume)
            args.start_epoch = checkpoint['epoch']
            best_prec1 = checkpoint['best_prec1']
            model.load_state_dict(checkpoint['state_dict'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.evaluate, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])

    train_loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(root='/media/data4b/haleyso/cifardata', train=True, transform=transforms.Compose([
            transforms.RandomHorizontalFlip(),
            transforms.RandomCrop(32, 4),
            transforms.ToTensor(),
            normalize,
        ]), download=True),
        batch_size=args.batch_size, shuffle=True,
        num_workers=args.workers, pin_memory=True, drop_last=True)

    if args.parameter_count:
        count_parameters(model)
        images, labels = next(iter(train_loader))
        print("image size", images[0].size())
        sys.exit("parameter count finished")

    val_loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(root='/media/data4b/haleyso/cifardata', train=False, transform=transforms.Compose([
            transforms.ToTensor(),
            normalize,
        ])),
        batch_size=args.batch_size, shuffle=False,
        num_workers=args.workers, pin_memory=True, drop_last=True)

    # define loss function (criterion) and pptimizer
    # criterion = nn.CrossEntropyLoss().cuda()
    criterion = losses.__dict__[args.loss]()
    print("criterion haley", criterion)
    # sys.exit("test")

    if args.half:
        model.half()
        criterion.half()

    optimizer = torch.optim.SGD(model.parameters(), args.lr,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay)

    lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, args.epochs, eta_min=0, last_epoch=-1)

    T_min, T_max = 1e-1, 1e1

    def Log_UP(K_min, K_max, epoch):
        Kmin, Kmax = math.log(K_min) / math.log(10), math.log(K_max) / math.log(10)
        return torch.tensor([math.pow(10, Kmin + (Kmax - Kmin) / args.epochs * epoch)]).float().cuda()

    def tau_iter(iter):
        new_tau = max(args.tau_min, args.tau_max * math.exp(-args.r * iter))
        return new_tau

    print(model.module)
    # sys.exit("test0")
    for epoch in range(args.start_epoch, args.epochs):

        if args.method == 'gumbel-softmax':
            tau = tau_iter(epoch)
            model.module.conv1.tau = tau
            model.module.conv2.tau = tau
            model.module.conv1.tau_max = args.tau_max
            model.module.conv2.tau_max = args.tau_max
            model.module.conv1.c = args.c_scale
            model.module.conv2.c = args.c_scale

            print("tau:", tau)
            print("max:", model.module.conv1.weight.max(), "min:", model.module.conv1.weight.min())
            if epoch >0:
                print("Quantized weights unique values", torch.unique(model.module.conv1.q_weights))

        elif args.method =='irnet':
            t = Log_UP(T_min, T_max, epoch)
            if (t < 1):
                k = 1 / t
            else:
                k = torch.tensor([1]).float().cuda()
            model.module.conv1.k = k
            model.module.conv2.k = k
            model.module.conv1.t = t
            model.module.conv2.t = t
        # train for one epoch
        print('current lr {:.5e}'.format(optimizer.param_groups[0]['lr']))
        train(train_loader, model, criterion, optimizer, epoch)

        # evaluate on validation set
        prec1 = validate(val_loader, model, criterion)
        lr_scheduler.step()

        # remember best prec@1 and save checkpoint
        is_best = prec1 > best_prec1
        best_prec1 = max(prec1, best_prec1)

        if epoch > 0 and epoch % args.save_every == 0:
            save_checkpoint({
                'epoch': epoch + 1,
                'state_dict': model.state_dict(),
                'best_prec1': best_prec1,
            }, is_best, filename=os.path.join(save_runs_dir, 'checkpoint.th'))
            # images, labels = next(iter(train_loader))
            # print(summary(model, images[0].size()), file=txtfile)
        if epoch == 0:
            images, labels = next(iter(train_loader))
            print(summary(model, images[0].size()))

    save_checkpoint({
        'state_dict': model.state_dict(),
        'best_prec1': best_prec1,
    }, is_best, filename=os.path.join(save_runs_dir, 'model.th'))

    # txtfile.close()



def train(train_loader, model, criterion, optimizer, epoch):
    """
        Run one train epoch
    """
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()

    # switch to train mode
    model.train()

    end = time.time()
    for i, (input, target) in enumerate(train_loader):

        # measure data loading time
        data_time.update(time.time() - end)

        target = target.cuda()
        input_var = torch.autograd.Variable(input).cuda()
        target_var = torch.autograd.Variable(target)
        if args.half:
            input_var = input_var.half()

        # compute output
        output = model(input_var)
        loss = criterion(output, target_var)

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        output = output.float()
        loss = loss.float()
        # measure accuracy and record loss
        prec1 = accuracy(output.data, target)[0]
        # losses.update(loss.data[0], input.size(0))
        # print("loss", loss, loss.data)
        losses.update(loss.item(), input.size(0))
        top1.update(prec1.item(), input.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            print('Epoch: [{0}][{1}/{2}]\t'
                  'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Data {data_time.val:.3f} ({data_time.avg:.3f})\t'
                  'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
                  'Prec@1 {top1.val:.3f} ({top1.avg:.3f})'.format(
                epoch, i, len(train_loader), batch_time=batch_time,
                data_time=data_time, loss=losses, top1=top1))


def validate(val_loader, model, criterion):
    """
    Run evaluation
    """
    batch_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()

    # switch to evaluate mode
    model.eval()

    end = time.time()
    for i, (input, target) in enumerate(val_loader):
        target = target.cuda()
        input_var = torch.autograd.Variable(input, volatile=True).cuda()
        target_var = torch.autograd.Variable(target, volatile=True)

        if args.half:
            input_var = input_var.half()

        # compute output
        output = model(input_var)
        loss = criterion(output, target_var)

        output = output.float()
        loss = loss.float()

        # measure accuracy and record loss
        prec1 = accuracy(output.data, target)[0]
        losses.update(loss.item(), input.size(0))
        top1.update(prec1.item(), input.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            print('Test: [{0}/{1}]\t'
                  'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
                  'Prec@1 {top1.val:.3f} ({top1.avg:.3f})'.format(
                i, len(val_loader), batch_time=batch_time, loss=losses,
                top1=top1))

    print(' * Prec@1 {top1.avg:.3f}'
          .format(top1=top1))

    return top1.avg


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    """
    Save the training model
    """
    torch.save(state, filename)
    pass


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

def count_parameters(model):
    table = PrettyTable(["Modules", "Parameters"])
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad: continue
        params = parameter.numel()
        table.add_row([name, params])
        total_params+=params
    print(table)
    print(f"Total Trainable Params: {total_params}")
    return total_params

if __name__ == '__main__':
    main()